/*

*/

package renderer.models;
import  renderer.scene.*;

/**
   Create a wireframe model of a frustum of a right square pyramid
   with its base in the xz-plane.
<p>
   See <a href="https://en.wikipedia.org/wiki/Frustum" target="_top">
                https://en.wikipedia.org/wiki/Frustum</a>

   @see Pyramid
*/
public class PyramidFrustum extends Model
{
   /**
      Create a frustum of a right square pyramid with its base in the
      xz-plane, a base side length of 2, top side length of 1, and height 1/2.
   */
   public PyramidFrustum( )
   {
      this(2.0, 1.0, 0.5, 7, 4);
   }


   /**
      Create a frustum of a right square pyramid with its base in the
      xz-plane, a base side length of {@code s1}, top side length of
      {@code s2}, and height {@code h}.
   <p>
      Notice that this model works with either {@code s1 > s2} or
      {@code s1 < s2}. In other words, the frustum can have its "apex"
      either above or below the xz-plane.

      @param s1  side length of the base of the frustum
      @param s2  side length of the top of the frustum
      @param h   height of the frustum
      @param n   number of lines of latitude
      @param k   number of lines of longitude
   */
   public PyramidFrustum(double s1, double s2, double h, int n, int k)
   {
      super();

      if (n < 0) n = 0;
      if (k < 1) k = 1;

      // Create the frustum's geometry.
      int index = 0;

      // Create all the lines of longitude from the top, down to the base,
      // across the base, then back up to the top, and across the top.
      s1 = s1/2;
      s2 = s2/2;
      double delta1 = (2 * s1) / k;
      double delta2 = (2 * s2) / k;
      // lines of "longitude" perpendicular to the x-axis
      for (int j = 0; j <= k; j++)
      {
         double d1 = j * delta1;
         double d2 = j * delta2;
         addVertex(new Vertex(-s2+d2, h, -s2));
         addVertex(new Vertex(-s1+d1, 0, -s1));
         addVertex(new Vertex(-s1+d1, 0,  s1));
         addVertex(new Vertex(-s2+d2, h,  s2));
         addLineSegment(index+0, index+1);
         addLineSegment(index+1, index+2);
         addLineSegment(index+2, index+3);
         addLineSegment(index+3, index+0);
         index += 4;
      }
      // lines of "longitude" perpendicular to the z-axis
      for (int j = 0; j <= k; j++)
      {
         double d1 = j * delta1;
         double d2 = j * delta2;
         addVertex(new Vertex( s2, h, -s2+d2));
         addVertex(new Vertex( s1, 0, -s1+d1));
         addVertex(new Vertex(-s1, 0, -s1+d1));
         addVertex(new Vertex(-s2, h, -s2+d2));
         addLineSegment(index+0, index+1);
         addLineSegment(index+1, index+2);
         addLineSegment(index+2, index+3);
         addLineSegment(index+3, index+0);
         index += 4;
      }
      // Create all the lines of "latitude" around the pyramid, starting
      // from the base and working up to the top.
      double deltaH = h / (n + 1);
      double deltaS = (s1 - s2) / (n + 1);
      double s = s1;
      for (int i = 0; i <= n; i++)
      {
         h = i * deltaH;
         addVertex(new Vertex( s, h,  s));
         addVertex(new Vertex( s, h, -s));
         addVertex(new Vertex(-s, h, -s));
         addVertex(new Vertex(-s, h,  s));
         addLineSegment(index+0, index+1);
         addLineSegment(index+1, index+2);
         addLineSegment(index+2, index+3);
         addLineSegment(index+3, index+0);
         s -= deltaS;
         index += 4;
      }
   }
}//PyramidFrustum
