/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer} data structure. This renderer mutates
   the {@link FrameBuffer}'s current viewport so that it is filled
   in with the rendered image of the scene represented by the
   {@link Scene} object.
<p>
   This implements our eigth rendering pipeline. This renderer
   implements "hierarchical scenes" by recursively traversing the
   DAG of nested {@link Position}s below each of a {@link Scene}'s
   {@link Position} objects. As the renderer traverses deeper into
   the DAG of nested {@link Position}s, it accumulates a "current
   transformation matrix" that transforms each {@link Vertex} from a
   {@link Model}'s local coordinate system to the {@link Camera}'s
   (shared) view coordinate system. The recursive traversal of the
   scene''s DAG is not a new pipeline stage, so there are still five
   pipeline stages.
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         if ( position.visible )
         {
            // Begin a pre-order, depth-first-traversal from this Position.
            render_position(scene, position, Matrix.identity(), fb);
         }
      }
   }


   /**
      Recursively renderer a {@link Position} without any debugging information.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then renders the {@link Model} in {@code position}
      using the updated {@code ctm} in the {@link Model2View} stage.


      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 0. Make a deep copy of the Position's Model.
         Model model2 = new Model(position.model);

         // 1. Apply the current model-to-view coordinate transformation.
         Model2View.model2view(model2.vertexList, ctm);

         // 2. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(model2.vertexList, scene.camera);

         // 3. Apply the projection transformation.
         Projection.project(model2.vertexList, scene.camera);

         // 4. Clip each line segment to the camera's view rectangle.
         List<LineSegment> lineSegmentList2 = new ArrayList<>();
         for (LineSegment ls : model2.lineSegmentList)
         {
            if ( Clip.clip(model2, ls) )
            {
               // Keep the line segments that are visible.
               lineSegmentList2.add(ls);
            }
         }
         model2.lineSegmentList = lineSegmentList2;

         // 5. Rasterize each visible line segment into pixels.
         for (LineSegment ls : model2.lineSegmentList)
         {
            RasterizeAntialias.rasterize(model2, ls, fb, doAntialiasing);
         }
      }

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order, depth-first-traversal from this nested Position.
            render_position(scene, p, ctm, fb);
         }
      }
   }



   /**
      Render the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         System.out.println("=== Render Position ================================");
         if ( position.visible )
         {
            // Begin a (pre-order) depth-first-traversal from this Position.
            render_position_debug(scene, position, Matrix.identity(), fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("=== End Position ================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a {@link Position} with quite a bit of debugging information.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then renders the {@link Model} in {@code position}
      using the updated {@code ctm} in the {@link Model2View} stage.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position_debug(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      System.out.println("=== Render Model ===================================");

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 0. Make a deep copy of the Position's Model.
         Model model2 = new Model(position.model);

         // log interesting information to standard output
         logVertexList("0. Input    ", model2.vertexList);

         // 1. Apply the current model-to-view coordinate transformation.
         Model2View.model2view(model2.vertexList, ctm);

         // log interesting information to standard output
         logVertexList("1. View     ", model2.vertexList);

         // 2. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(model2.vertexList, scene.camera);

         // log interesting information to standard output
         logVertexList("2. Camera   ", model2.vertexList);

         // 3. Apply the projection transformation.
         Projection.project(model2.vertexList, scene.camera);

         // log interesting information to standard output
         logVertexList("3. Projected", model2.vertexList);

         // 4. Clip each line segment to the camera's view rectangle.
         List<LineSegment> lineSegmentList2 = new ArrayList<>();
         for (LineSegment ls : model2.lineSegmentList)
         {
            // log interesting information to standard output
            logLineSegment("4. Clipping", model2, ls);

            if ( Clip.clip(model2, ls) )
            {
               // log interesting information to standard output
               logLineSegment("4. Clipped (accept)", model2, ls);

               // Keep the line segments that are visible.
               lineSegmentList2.add(ls);
            }
            else
            {
               // log interesting information to standard output
               logLineSegment("4. Clipped (reject)", model2, ls);
            }
         }
         model2.lineSegmentList = lineSegmentList2;

         // log interesting information to standard output
         logVertexList("3. Clipped  ", model2.vertexList);

         // 5. Rasterize each visible line segment into pixels.
         for (LineSegment ls : model2.lineSegmentList)
         {
            // log interesting information to standard output
            logLineSegment("5. Rasterize", model2, ls);

            RasterizeAntialias.rasterize(model2, ls, fb, doAntialiasing);
         }
         System.out.println();
      }
      else
      {
         System.out.println("   Hidden model.");
      }

      System.out.println("=== End Model ===================================");

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         System.out.println("====== Render Nested Position ========================");
         if ( p.visible )
         {
            // Do a (pre-order) depth-first-traversal from this nested Position.
            render_position_debug(scene, p, ctm, fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("====== End Nested Position ===========================");
      }
   }


   /**
      This method prints a {@link String} representation of the given vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, Model model, LineSegment ls)
   {
      System.out.print( stage + ": " + ls.toString() );
      int vIndex0 = ls.vIndex[0];
      int vIndex1 = ls.vIndex[1];
      Vertex v0 = model.vertexList.get(vIndex0);
      Vertex v1 = model.vertexList.get(vIndex1);
      int cIndex0 = ls.cIndex[0];
      int cIndex1 = ls.cIndex[1];
      Color c0 = model.colorList.get(cIndex0);
      Color c1 = model.colorList.get(cIndex1);
      System.out.printf("   vIndex = %3d, " + v0.toString(), vIndex0);
      System.out.printf("   vIndex = %3d, " + v1.toString(), vIndex1);
      System.out.printf("   cIndex = %3d, " + c0.toString() + "\n", cIndex0);
      System.out.printf("   cIndex = %3d, " + c1.toString() + "\n", cIndex1);
   }
}
